﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace FormExample
{
    /// <summary>
    /// Simple program that allows you to verify the presence of a valid
    /// SmartDongle, read data from the SmartDongle, write data to the
    /// SmartDongle, check the SmartDongle's serial number, and play with
    /// the SmartDongle's LED.
    /// 
    /// The "Check Valid" button will check if a valid SmartDongle is
    /// connected to the computer and return any error messages.
    /// 
    /// The "Read Data" button will read data from the SmartDongle based
    /// on the given address and number of bytes to read.
    /// 
    /// The "Write Data" button will write the text from the "Data Read
    /// From SmartDongle / Data to Write to SmartDongle" box to the
    /// SmartDongle. It will then show the number of bytes/characters
    /// written in the "Number of Bytes to Read/Written" box. 
    /// 
    /// The "SmartDongle Messages" box displays error messages returned
    /// from the SmartDongle.
    /// 
    /// The "Data Read From SmartDongle / Data to Write to SmartDongle"
    /// box displays the sting representation of data read from the
    /// SmartDongle. You can also enter text to be written to the
    /// SmartDongle here.
    /// 
    /// The "Address to Read/Write From/To" box allows you to enter the
    /// memory address on the SmartDongle which you would like to read
    /// from or write to. It must be a valid number between 0 and 32431
    /// as this is the available memory on the SmartDongle.
    /// 
    /// The "Number of Bytes to Read/Written" box lets you tell the
    /// program how many bytes you want to read from the SmartDongle
    /// and tells you how many bytes were written during a write
    /// operation. Its value must be less than or equal to: 
    ///     32431 - Address + 1.
    /// 
    /// The "LED Controls" allow you to turn the LED off or change
    /// its color to red or green.
    /// 
    /// The "Get SmartDongle Serial #" button will display the
    /// SmartDongle's serial number in the "S/N" box.
    ///
    /// MicroWorks, Inc.
    /// 2808 North Cole Road
    /// Boise, ID 83704
    /// 208-375-1234
    ///
    /// http://www.smartdongle.com
    ///
    /// </summary>

    public partial class formExample : Form
    {
        /* 
         * Your keys here
         * Leave the 0x at the beginning.
         */
        UInt64 P1 = 0xec6cc589aefd1e75;
        UInt64 P2 = 0xfcec0a6a82747b3f;

        public formExample()
        {
            InitializeComponent();
        }

        // Turn LED off.
        private void ledOff_CheckedChanged(object sender, EventArgs e)
        {
            SmartDongle.Off();
        }

        // Turn LED red.
        private void ledRed_CheckedChanged(object sender, EventArgs e)
        {
            SmartDongle.Red();
        }

        // Turn LED Green.
        private void ledGreen_CheckedChanged(object sender, EventArgs e)
        {
            SmartDongle.Green();
        }

        // Check the computer for a valid SmartDongle and returns
        // the associated error message.
        private void checkValid_Click(object sender, EventArgs e)
        {
            smartDongleData.Text = "";
            SmartDongle.Red();
            ledRed.Checked = true;
 
            // Reading the SmartDongle at address 0 with null buffer and 0 size
            // to check for existence instead of actually reading.
            SmartDongle.SmartDnglError error;
            UInt64 tempP1;
            UInt64 tempP2;
            DateTime start;
            if (!(diffKeys.Checked)) {
                start = DateTime.Now;
                error = SmartDongle.Read(P1, P2, 0, null, 0);
            }
            else {
                try {
                    tempP1 = Convert.ToUInt64(tempKey1.Text, 16);
                    tempP2 = Convert.ToUInt64(tempKey2.Text, 16);
                } catch (Exception ex) {
                    MessageBox.Show("Invalid Key Format.\n" + ex.ToString());
                    return;
                }
                start = DateTime.Now;
                error = SmartDongle.Read(tempP1, tempP2, 0, null, 0);
            }

            TimeSpan elapsed = DateTime.Now - start;
            SmartDongle.Green();
            ledGreen.Checked = true;
            if (error == 0) {
                if (timeFunctions.Checked)
                    smartDongleData.Text = "Check completed in " + elapsed.TotalSeconds + " sec";
                smartDongleMessages.Text = "Valid SmartDongle Found.";
            }
            else
                smartDongleMessages.Text = SmartDongle.GetErrorString(error);
        }

        // Read data from the SmartDongle based on the user's given
        // address and number of bytes.
        private void readData_Click(object sender, EventArgs e)
        {
            int dataError = 0;
            ushort addr = 0;
            int numBytes = 0;

            // Ensure the address is a valid number within range.
            try {
                addr = Convert.ToUInt16(readWriteAddress.Text);
                if (addr > 32431) {
                    smartDongleData.Text = "Overflow Exception: Address value must be between 0 and 32431";
                    dataError = 1;
                }
            }
            catch (FormatException) {
                smartDongleData.Text = "Format Exception: Address value must be a valid number between 0 and 32431";
                dataError = 1;
            }
            catch (OverflowException) {
                smartDongleData.Text = "Overflow Exception: Address value must be between 0 and 32431";
                dataError = 1;
            }

            // Ensure the number of bytes is a valid number within range.
            try {
                numBytes = Convert.ToInt32(numBytesRead.Text);
                if ((addr + numBytes - 1) > 32431) {
                    smartDongleData.Text = "Overflow Exception: (Address + Number of Bytes - 1) must be <= 32431";
                    dataError = 1;
                }
            }
            catch (FormatException) {
                smartDongleData.Text = "Format Exception: Number of Bytes value must be a valid number between 0 and 32431";
                dataError = 1;
            }
            catch (OverflowException) {
                smartDongleData.Text = "Overflow Exception: Number of Bytes value must be between 0 and 32431";
                dataError = 1;
            }

            // If address and number of bytes are valid.
            if (dataError == 0) {
                byte[] buff = new byte[numBytes];

                SmartDongle.Red();
                ledRed.Checked = true;
                SmartDongle.SmartDnglError error;
                UInt64 tempP1;
                UInt64 tempP2;
                DateTime start;

                // Read SmartDongle at given address for given number of bytes.
                if (!diffKeys.Checked) {
                    start = DateTime.Now;
                    error = SmartDongle.Read(P1, P2, addr, buff, numBytes);
                }
                else {
                    try {
                        tempP1 = Convert.ToUInt64(tempKey1.Text, 16);
                        tempP2 = Convert.ToUInt64(tempKey2.Text, 16);
                    } catch (Exception ex) {
                        MessageBox.Show("Invalid Key Format.\n" + ex.ToString());
                        return;
                    }
                    start = DateTime.Now;
                    error = SmartDongle.Read(tempP1, tempP2, addr, buff, numBytes);
                }

                TimeSpan elapsed = DateTime.Now - start;
                SmartDongle.Green();
                ledGreen.Checked = true;

                // If error reading the SmartDongle, return error message.
                if (error != 0) {
                    smartDongleMessages.Text = SmartDongle.GetErrorString(error);
                    smartDongleData.Text = "";
                }
                // Else if no error, return data read from SmartDongle.
                else {
                    smartDongleMessages.Text = "";
                    if (timeFunctions.Checked)
                        smartDongleData.Text = "Read completed in " + elapsed.TotalSeconds + " sec\n" + System.Text.Encoding.UTF8.GetString(buff);
                    else
                        smartDongleData.Text = System.Text.Encoding.UTF8.GetString(buff);
                }
            }
        }
       /*
        * Write data (given by the user) to the address (given by the user)
        * on the SmartDongle. It then reads the data back from the SmartDongle
        * and compares what was read to what should have been written and
        * displays a match/not match note along with the data read from the
        * SmartDongle.
       */ 
        private void writeData_Click(object sender, EventArgs e)
        {
            int dataError = 0;
            ushort addr = 0;
            // Ensure the entered address is a valid number within range.
            try {
                addr = Convert.ToUInt16(readWriteAddress.Text);
                if (addr > 32431) {
                    smartDongleData.Text = "Overflow Exception: Address value must be between 0 and 32431";
                    dataError = 1;
                }
            }
            catch (FormatException) {
                smartDongleData.Text = "Format Exception: Address value must be a valid number between 0 and 32431";
                dataError = 1;
            }
            catch (OverflowException) {
                smartDongleData.Text = "Overflow Exception: Address value must be between 0 and 32431";
                dataError = 1;
            }

            // Get the number of bytes to be written and ensure
            // it is not too large for the given address.
            int numBytes = smartDongleData.Text.Length;
            if (addr + numBytes - 1 > 32431) {
                smartDongleData.Text = "Overflow Exception: (Address + Number of Bytes (characters) to Write - 1) must be <= 32431";
                dataError = 1;
            }

            // If address and number of bytes are valid.
            if (dataError == 0) {
                // Copy data to write to a byte buffer.
                byte[] writeBuff = System.Text.Encoding.UTF8.GetBytes(smartDongleData.Text);

                byte[] readBuff = new byte[numBytes];

                SmartDongle.Red();
                ledRed.Checked = true;
                SmartDongle.SmartDnglError error;
                UInt64 tempP1 = 0;
                UInt64 tempP2 = 0;
                DateTime start;
                
                // Write the given data to the given address on the SmartDongle.
                if (!diffKeys.Checked) {
                    start = DateTime.Now;
                    error = SmartDongle.Write(P1, P2, addr, writeBuff, numBytes);
                }
                else {
                    try {
                        tempP1 = Convert.ToUInt64(tempKey1.Text, 16);
                        tempP2 = Convert.ToUInt64(tempKey2.Text, 16);
                    }
                    catch (Exception ex) {
                        MessageBox.Show("Invalid Key Format.\n" + ex.ToString());
                        return;
                    }
                    start = DateTime.Now;
                    error = SmartDongle.Write(tempP1, tempP2, addr, writeBuff, numBytes);
                }

                TimeSpan elapsed = DateTime.Now - start;
                SmartDongle.Green();
                ledGreen.Checked = true;

                // If there was an error writing to the SmartDongle,
                // return the error message.
                if (error != 0) {
                    smartDongleMessages.Text = SmartDongle.GetErrorString(error);
                    smartDongleData.Text = "";
                }
               /*
                * Else if there was no write error, read the data back
                * from the SmartDongle and check if it matches what was
                * supposed to be written.
               */ 
                else {
                    smartDongleMessages.Text = "";

                    SmartDongle.Red();
                    ledRed.Checked = true;

                    // Read the SmartDongle where the data was just written to.
                    if (!diffKeys.Checked)
                        error = SmartDongle.Read(P1, P2, addr, readBuff, numBytes);
                    else
                        error = SmartDongle.Read(tempP1, tempP2, addr, readBuff, numBytes);

                    SmartDongle.Green();

                    // If there is an error reading the SmartDongle, return
                    // the error message.
                    ledGreen.Checked = true;
                    if (error != 0) {
                        smartDongleMessages.Text = SmartDongle.GetErrorString(error);
                        smartDongleData.Text = "";
                    }
                   /*
                    * Else if no error, compare the written and read string and
                    * return a match/no match message along with the data that
                    * was read from the SmartDongle.
                   */ 
                    else {
                        smartDongleMessages.Text = "";
                        string writeString = System.Text.Encoding.UTF8.GetString(writeBuff);
                        string readString = System.Text.Encoding.UTF8.GetString(readBuff);
                        if (writeString == readString)
                            readString = "Data Written Matches Data Read.\nData Read:\n" + readString;
                        else
                            readString = "Data Written Does Not Match Data Read.\nData Read:\n" + readString;
                        numBytesRead.Text = numBytes.ToString();
                        if (timeFunctions.Checked)
                            readString = "Data written in " + elapsed.TotalSeconds + " sec\n" + readString;
                        smartDongleData.Text = readString;
                    }
                }
            }
        }

        // Display the SmartDongle's serial number.
        private void getSerialNumber_Click(object sender, EventArgs e)
        {
            string sn = "";
            SmartDongle.GetSerialNumber(ref sn);
            serialNumber.Text = sn;
        }
    }
}
